# SPDX-License-Identifier: MIT

ALLOWLIST="/etc/atmark/usb-filter-allowlist"
DEVICE_LIST="/run/usb-filter-device-list"
CLASS_LIST="/usr/share/abos-ctrl/usb_device_class"
UBOOT_ENV="/boot/uboot_env.d/80_optargs"

usb_filter_help() {
	echo "Usage: abos-ctrl usb-filter [action [arguments]]"
	echo
	echo "Possible actions:"
	echo "  enable: enabling the USB filter function"
	echo "  disable: disabling the USB filter"
	echo "  list-devices: list currently connected USB devices"
	echo "  list-rules: list currently USB device allowlist"
	echo "  reset-rules [--force]: reset USB device allowlist"
	echo "                         with --force option, do not prompt"
	echo "  allow-device {ID}: allow connected USB devices specified by ID to connect,"
	echo "                     and add device's information to the allowlist"
	echo "                     ID can be checked with \`abos-ctrl usb-filter list-devices\`"
	echo "  block-device {ID}: block alllowed USB devices specified by ID to connect,"
	echo "                     and delete device's information from the allowlist"
	echo "                     ID can be checked with \`abos-ctrl usb-filter list-devices\`"
	echo "  allow-class [CLASS]: create allow rule for each USB device class"
	echo "                       the string for 'CLASS' can be found by running"
	echo "                       \`abos-ctrl usb-filter allow-class\`"
	echo "  remove-rule {ID}: remove the allow rule specified by ID"
	echo "                    ID can be checked with \`abos-ctrl usb-filter list-rules\`"
	echo "  help: show this message"
}

usb_filter_set_auth_default() {
	value="$1"
	[ "$value" != "0" ] && [ "$value" != "1" ] && \
		error "usb_filter_set_auth_default: " \
			  "Only '0' or '1' can be specified as an arguments"

	for auth_def in /sys/bus/usb/devices/usb*/authorized_default; do
		echo "$value" > "$auth_def" || exit 1
	done

	local optargs
	optargs="$(fw_printenv | grep "^optargs=" \
			 | sed -e 's/usbcore.authorized_default=.//' \
			 -e 's/[[:space:]]\+$//')"
	optargs="$optargs usbcore.authorized_default=$value"
	echo "$optargs" > "$UBOOT_ENV" && \
		persist_file "$UBOOT_ENV" && \
		fw_setenv_nowarn -s "$UBOOT_ENV" || exit 1
}

usb_filter_enable() {
	usb_filter_set_auth_default "0"
	echo "USB filter enabled."
	echo "please reboot to apply."
}

usb_filter_disable() {
	usb_filter_set_auth_default "1"
	echo "USB filter disabled."
	echo "please reboot to apply."
}

usb_filter_status() {
	local status="disabled"
	fw_printenv | grep -q "usbcore.authorized_default=0" && \
		status="enabled"
	echo "$status"
}

usb_filter_check_device_class() {
	local allowed_classes target_classes
	allowed_classes="$(awk '/^class/ {print $2}' "$ALLOWLIST")"
	[ -z "$allowed_classes" ] && return 1
	target_classes="$(echo "$ID_USB_INTERFACES" | \
awk -F ':' '{
	for (i=2; i<NF; i++) {
		{ print substr($i, 0, 2) }
	}
}' | uniq)"

	for c in $target_classes; do
		local target_class
		target_class="$(awk -F '=' -v c="$c" '$2 == c {print $1}' "$CLASS_LIST")"
		[ -z  "$target_class" ] && return 1
		grep -q "$target_class" "$ALLOWLIST" 2>/dev/null || return 1
	done

	return 0
}

usb_filter_authorize_device() {
	if usb_filter_check_device_class; then
		echo 1 > /sys"$DEVPATH"/authorized || exit 1
		return
	fi

	local device index=2
	device="$(cat $ALLOWLIST)"
	for param in "$ID_VENDOR_ID" "$ID_MODEL_ID" "$ID_MODEL" \
				 "$ID_USB_INTERFACES" "$ID_SERIAL"; do
		device="$(echo "$device" | awk \
				-v param="\"$param\"" -v col="$index" \
				'/^device/ && ($col == param || $col == "\"*\"") {print}')"
		[ -z "$device" ] && return
		index=$((index + 1))
	done

	echo 1 > /sys"$DEVPATH"/authorized || exit 1
}

usb_filter_add_device() {
	[ "$ACTION" != "add" ] && \
		error "This option can only be executed from udev rules"

	[ "$(cat "/sys$DEVPATH/devpath")" = "0" ] && return

	[ -n "$BUSNUM" ] || return
	[ -n "$DEVNUM" ] || return

	for param in "$ID_VENDOR_ID" "$ID_MODEL_ID" "$ID_MODEL" \
				 "$ID_USB_INTERFACES" "$ID_SERIAL"; do
		echo "$param" | grep -q " " && \
			error "recognised USB device information contains white space"
	done

	# usbinfo format
	# BUSNUM:DEVNUM,VENDOR_ID,MODEL_ID,MODEL_NAME,USB_INTERFACES,SERIAL,DEVPATH
	local usbinfo=
	usbinfo="\"$BUSNUM:$DEVNUM\""
	usbinfo="$usbinfo \"$ID_VENDOR_ID\""
	usbinfo="$usbinfo \"$ID_MODEL_ID\""
	usbinfo="$usbinfo \"$ID_MODEL\""
	usbinfo="$usbinfo \"$ID_USB_INTERFACES\""
	usbinfo="$usbinfo \"$ID_SERIAL\""
	usbinfo="$usbinfo \"$DEVPATH\""
	echo "$usbinfo" >> $DEVICE_LIST || exit 1

	[ "$(cat /sys"$DEVPATH"/authorized)" != 0 ] && exit
	usb_filter_authorize_device
}

usb_filter_remove_device() {
	[ "$ACTION" != "remove" ] && \
		error "This option can only be executed from udev rules"

	[ -n "$BUSNUM" ] || return
	[ -n "$DEVNUM" ] || return
	sed -i -e "/^\"$BUSNUM:$DEVNUM\"/d" $DEVICE_LIST
}

usb_filter_allow_device() {
	local id=$1 devinfo
	target="$(awk -v id="$id" 'NR==id' $DEVICE_LIST)"
	[ -z "$target" ] && \
		error "No device with such an ID exists"
	devinfo="$(echo "$target" | awk '{$1=""; print}')"

	local devpath allow_list
	devpath="$(echo "$devinfo" | awk '{gsub("^\"|\"$", "", $6); print $6}')"
	echo 1 > /sys"$devpath"/authorized || exit 1
	echo "$devinfo" | \
		awk '{$6=""; print "device " $0}' >> "$ALLOWLIST" || \
		exit 1
	allow_list="$(uniq $ALLOWLIST)"
	echo "$allow_list" > "$ALLOWLIST" || exit 1
	persist_file "$ALLOWLIST"
	echo "allowed the following device:"
	echo "    $target"
}

usb_filter_block_device() {
	local id=$1

	if [ "$(usb_filter_status)" != "enabled" ]; then
		error "USB filter is currently disabled" \
			  "Run \`abos-ctrl usb-filter enable\` first to enable USB filtering"
		return
	fi

	local devinfo
	devinfo="$(awk -v id="$id" 'NR==id' $DEVICE_LIST)"
	[ -z "$devinfo" ] && \
		error "No device with such an ID exists"

	local serial devpath
	serial="$(echo "$devinfo" | awk '{print $6}')"
	sed -i -e "/$serial/d" $ALLOWLIST 2>/dev/null
	devpath="$(echo "$devinfo" | awk '{gsub("^\"|\"$", "", $7); print $7}')"
	[ "$(cat /sys"$devpath"/authorized)" = "0" ] && return

	# When removing an authorised USB device,
	# it is necessary to run udevadm trigger to correctly remove
	# the associated symlink and information about the device.
	udevadm trigger --action=remove /sys"$devpath"
	echo 0 > /sys"$devpath"/authorized || exit 1
	udevadm trigger --action=add /sys"$devpath"

	echo "blocked the following device:"
	echo "    $devinfo"
}

usb_filter_allow_class() {
	local class=$1
	if [ -z "$class" ]; then
		echo "supported USB device classes:"
		awk -F '=' '{print "\t"$1}' "$CLASS_LIST"
		return
	fi

	local id allowlist
	id="$(awk -F '=' -v class="$class" \
		'$1 == class {print $2}' "$CLASS_LIST")"
	if [ -z "$id" ]; then
		error "$class is wrong argument." \
			  "supported USB device classes:" \
			  "$(awk -F '=' '{print "\t"$1}' "$CLASS_LIST")"
	fi
	echo "class $class" >> "$ALLOWLIST" || exit 1
	allowlist="$(sort $ALLOWLIST | uniq)"
	echo "$allowlist" > "$ALLOWLIST" || exit 1
	persist_file "$ALLOWLIST"
}

usb_filter_list_devices() {
	if [ ! -s "$DEVICE_LIST" ]; then
		echo "No USB devices connected"
		return 1
	fi

	local index=1
	local delete_list=""
	while read -r line
	do
		local devpath status="block"
		devpath="$(echo "$line" | awk '{gsub("\"", "", $7); print $7}')"
		if [ ! -e "/sys$devpath/authorized" ]; then
			warning "/sys$devpath/authorized is not found." \
					"This device may not already be connected."
			delete_list="$(echo "$line" | awk '{print $1}') $delete_list"
			continue
		fi
		[ "$(cat /sys"$devpath"/authorized)" = "1" ] && status="allow"
		echo "     $index  $status $line"
		index=$((index + 1))
	done < "$DEVICE_LIST"

	for delete in $delete_list; do
		sed -i -e "/^$delete/d" "$DEVICE_LIST"
	done
}

usb_filter_list_rules() {
	if [ ! -s "$ALLOWLIST" ]; then
		echo "No rules have been set"
		return 1
	fi
	cat -n "$ALLOWLIST"
}

usb_filter_remove_rule() {
	local id=$1 target
	target="$(cat $ALLOWLIST | awk -v id="$id" 'NR==id')"
	[ -z "$target" ] && error "No rule with such an ID exists"

	sed -i -e "${id}d" "$ALLOWLIST"
	persist_file "$ALLOWLIST"
}

usb_filter_reset_rules() {
	if [ "$1" = "--force" ]; then
		persist_file -d "$ALLOWLIST"
		echo "All rules have been removed."
		return
	fi

	if usb_filter_list_rules; then
		echo ""
		echo "This command will delete all the above configured rules."
		echo "Continue? [y/N]"
		prompt_yesno n || error "Canceled."

		persist_file -d "$ALLOWLIST"
		echo "All rules have been removed."
	else
		echo "so nothing to do."
		return
	fi
}

ctrl_usb_filter() {
	if [ "$#" -eq 0 ]; then
		echo "Currently USB filter is $(usb_filter_status)."
		return
	fi

	case "$1" in
	"add")
		usb_filter_add_device
		;;
	"remove")
		usb_filter_remove_device
		;;
	"allow-device")
		[ "$#" -eq 2 ] || error "This command requires one argument."
		usb_filter_allow_device "$2"
		;;
	"block-device")
		[ "$#" -eq 2 ] || error "This command requires one argument."
		usb_filter_block_device "$2"
		;;
	"allow-class")
		usb_filter_allow_class "$2"
		;;
	"list-devices")
		usb_filter_list_devices
		;;
	"reset-rules")
		usb_filter_reset_rules "$2"
		;;
	"remove-rule")
		[ "$#" -eq 2 ] || error "This command requires one argument."
		usb_filter_remove_rule "$2"
		;;
	"list-rules")
		usb_filter_list_rules
		;;
	"enable")
		usb_filter_enable
		;;
	"disable")
		usb_filter_disable
		;;
	"help")
		usb_filter_help
		;;
	*)
		warning "Invalid option $1"
		usb_filter_help >&2
		exit 1
		;;
	esac
}
